using System;
using System.Numerics;
using System.Runtime.InteropServices;
using System.Text;
using SDL2;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.Graphics;
using SlatedGameToolkit.Framework.Graphics.Window;

namespace SlatedGameToolkit.Framework.Graphics.Programs
{
    /// <summary>
    /// Represents a GL Program.
    /// This class is abstract and serves as a parent to the possible types of programs.
    /// Gives access to common OpenGL program associated function calls.
    /// </summary>
    public abstract class GLProgram : IDisposable
    {
        private WindowContext context;
        protected readonly UIntPtr handle;
        protected GLCreateProgram createProgram;
        protected GLDeleteProgram deleteProgram;
        protected GLLinkProgram linkProgram;
        protected GLGetProgramInfoLog getProgramInfoLog;
        protected GLProgramParameter programParameter;
        private GLUseProgram useProgram;
        private readonly GLGetAttribLocation getAttribLocation;
        private readonly GLGetUniformLocation getUniformLocation;
        private readonly GLUniformMatrix4fv uniformMatrix4Fv;


        /// <summary>
        /// Creates an OpenGL program.
        /// 
        /// This is bound to the current context.
        /// </summary>
        internal GLProgram() {
            this.context = WindowContextsManager.CurrentWindowContext;
            createProgram = OpenGL.RetrieveGLDelegate<GLCreateProgram>("glCreateProgram");
            deleteProgram = OpenGL.RetrieveGLDelegate<GLDeleteProgram>("glDeleteProgram");
            linkProgram = OpenGL.RetrieveGLDelegate<GLLinkProgram>("glLinkProgram");
            getProgramInfoLog = OpenGL.RetrieveGLDelegate<GLGetProgramInfoLog>("glGetProgramInfoLog");
            programParameter = OpenGL.RetrieveGLDelegate<GLProgramParameter>("glProgramParameteri");
            useProgram = OpenGL.RetrieveGLDelegate<GLUseProgram>("glUseProgram");
            getAttribLocation = OpenGL.RetrieveGLDelegate<GLGetAttribLocation>("glGetAttribLocation");
            getUniformLocation = OpenGL.RetrieveGLDelegate<GLGetUniformLocation>("glGetUniformLocation");
            uniformMatrix4Fv = OpenGL.RetrieveGLDelegate<GLUniformMatrix4fv>("glUniformMatrix4fv");

            handle = createProgram();
            OpenGLErrorException.CheckGLErrorStatus();
        }


        public int GetAttributeLocation(string name) {
            int res = getAttribLocation(handle, name);
            OpenGLErrorException.CheckGLErrorStatus();
            return res;
        }

        public int GetUniformLocation(string name) {
            int res = getUniformLocation(handle, name);
            OpenGLErrorException.CheckGLErrorStatus();
            return res;
        }

        public void SetUniformMatrix4x4(int location, Matrix4x4 matrix, bool transpose = false) {
            Use();
            float[] mat4 = new float[] {
                matrix.M11, matrix.M21, matrix.M31, matrix.M41,
                matrix.M21, matrix.M22, matrix.M23, matrix.M24,
                matrix.M31, matrix.M32, matrix.M33, matrix.M34,
                matrix.M41, matrix.M42, matrix.M43, matrix.M44
            };
            uniformMatrix4Fv(location, 1, transpose, mat4);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        /// <summary>
        /// Binds this program.
        /// </summary>
        public virtual void Use() {
            useProgram(handle);
            uint written;
            byte[] log = new byte[2048];
            getProgramInfoLog(handle, (uint)log.Length, out written, log);
            if (written > 0) {
                throw new OpenGLException(Encoding.UTF8.GetString(log));
            }
            OpenGLErrorException.CheckGLErrorStatus();
        }

        /// <summary>
        /// Disposes of the program.
        /// </summary>
        public virtual void Dispose()
        {
            deleteProgram(handle);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        ~GLProgram() {
            Dispose();
        }
    }
}