using System;
using System.Runtime.InteropServices;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.Graphics;

namespace SlatedGameToolkit.Framework.Graphics.Shaders
{
    public sealed class GLVertexArraySet : IDisposable {
        private bool disposed;
        public int Count { get; private set; }
        private UIntPtr[] arrayBufferHandles;
        private UIntPtr[] vertexArrayHandles;
        private UIntPtr[] indexArrayHandles;
        private GLBindBuffer bindBuffer;
        private GLGenBuffers genBuffers;
        private GLDeleteBuffers deleteBuffers;
        private GLGenVertexArrays genVertexArrays;
        private GLBindVertexArray bindVertexArray;
        private GLDeleteVertexArrays deleteVertexArrays;
        private GLVertexAttribPointer vertexAttribPointer;
        private GLEnableVertexAttribArray enableVertexAttribArray;
        private GLDisableVertexAttribArray disableVertexAttribArray;
        private GLBufferData bufferData;
        public GLVertexArraySet(uint amount) {
            if (amount > int.MaxValue) throw new ArgumentOutOfRangeException("amount");
            this.Count = (int)amount;
            genBuffers = OpenGL.RetrieveGLDelegate<GLGenBuffers>("glGenBuffers");
            deleteBuffers = OpenGL.RetrieveGLDelegate<GLDeleteBuffers>("glDeleteBuffers");
            bindBuffer = OpenGL.RetrieveGLDelegate<GLBindBuffer>("glBindBuffer");
            genVertexArrays = OpenGL.RetrieveGLDelegate<GLGenVertexArrays>("glGenVertexArrays");
            bindVertexArray = OpenGL.RetrieveGLDelegate<GLBindVertexArray>("glBindVertexArray");
            deleteVertexArrays = OpenGL.RetrieveGLDelegate<GLDeleteVertexArrays>("glDeleteVertexArrays");
            vertexAttribPointer = OpenGL.RetrieveGLDelegate<GLVertexAttribPointer>("glVertexAttribPointer");
            enableVertexAttribArray = OpenGL.RetrieveGLDelegate<GLEnableVertexAttribArray>("glEnableVertexAttribArray");
            disableVertexAttribArray = OpenGL.RetrieveGLDelegate<GLDisableVertexAttribArray>("glDisableVertexAttribArray");
            bufferData = OpenGL.RetrieveGLDelegate<GLBufferData>("glBufferData");
            arrayBufferHandles = new UIntPtr[amount];
            vertexArrayHandles = new UIntPtr[amount];
            indexArrayHandles = new UIntPtr[amount];
            genBuffers(amount, arrayBufferHandles);
            OpenGLErrorException.CheckGLErrorStatus();
            genVertexArrays(amount, vertexArrayHandles);
            OpenGLErrorException.CheckGLErrorStatus();
            genBuffers(amount, indexArrayHandles);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        public void Use(uint index) {
            bindVertexArray(vertexArrayHandles[index]);
            OpenGLErrorException.CheckGLErrorStatus();
            bindBuffer(GLEnum.GL_ELEMENT_ARRAY_BUFFER, indexArrayHandles[index]);
            bindBuffer(GLEnum.GL_ARRAY_BUFFER, arrayBufferHandles[index]);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        public unsafe void BufferVertices(uint index, float[] data, bool dynamic) {
            Use(index);
            bufferData(GLEnum.GL_ARRAY_BUFFER, (uint) (sizeof(float) * data.Length), data, dynamic ? GLEnum.GL_DYNAMIC_DRAW : GLEnum.GL_STATIC_DRAW);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        public unsafe void BufferIndices(uint index, uint[] data, bool dynamic) {
            Use(index);
            bufferData(GLEnum.GL_ELEMENT_ARRAY_BUFFER, (uint) (sizeof(float) * data.Length), data, dynamic ? GLEnum.GL_DYNAMIC_DRAW : GLEnum.GL_STATIC_DRAW);
            OpenGLErrorException.CheckGLErrorStatus();
        }

        /// <summary>
        /// Defines the vertex attributes in an OpenGL vertex array object.
        /// Sends as floats.
        /// </summary>
        /// <param name="bufferIndex">The buffer in this set to affect.</param>
        /// <param name="enableAttributes">Whether or not to automatically enable the attributes that are being defined. Defaults to true.</param>
        /// <param name="definitions">The definitions for the vertex array object.</param>
        public void defineVertexAttributes(uint bufferIndex, bool enableAttributes = true, params VertexAttributeDefinition[] definitions) {
            Use(bufferIndex);
            foreach (VertexAttributeDefinition definition in definitions)
            {
                vertexAttribPointer(definition.attributeIndex, definition.size, GLEnum.GL_FLOAT, false, definition.stride, definition.offset);
                OpenGLErrorException.CheckGLErrorStatus();
                if (enableAttributes) enableVertexAttribArray(definition.attributeIndex);
            }
        }

        /// <summary>
        /// Enaables the vertex array object's definitions at the given attribute indices.
        /// </summary>
        /// <param name="bufferIndex">The buffer index in this set this should affect.</param>
        /// <param name="attributeIndices">The attribute indices to be enabled.</param>
        public void EnableAttributes(uint bufferIndex, params uint[] attributeIndices) {
            Use(bufferIndex);
            foreach (uint attrib in attributeIndices)
            {
                enableVertexAttribArray(attrib);
            }
        }

        /// <summary>
        /// Disables the vertex array object's definitions at the given attribute indices.
        /// </summary>
        /// <param name="bufferIndex">The buffer index in this set that should be affected.</param>
        /// <param name="attributeIndices">The attribute indices to be disabled.</param>
        public void DisableAttributes(uint bufferIndex, params uint[] attributeIndices) {
            Use(bufferIndex);
            foreach (uint attrib in attributeIndices)
            {
                disableVertexAttribArray(attrib);
            }
        }

        public void Dispose()
        {
            if (this.disposed) return;
            this.disposed = true;
            deleteBuffers((uint)arrayBufferHandles.Length, arrayBufferHandles);
            OpenGLErrorException.CheckGLErrorStatus();
            deleteVertexArrays((uint)vertexArrayHandles.Length, vertexArrayHandles);
            OpenGLErrorException.CheckGLErrorStatus();
            GC.SuppressFinalize(this);
        }

        ~GLVertexArraySet() {
            Dispose();
        }
    }

    public struct VertexAttributeDefinition {
        public uint attributeIndex;
        public int size;
        public uint stride;
        public uint offset;

        public VertexAttributeDefinition(uint attributeIndex, int size, uint stride, uint offset) {
            this.attributeIndex = attributeIndex;
            this.size = size;
            this.stride = stride;
            this.offset = offset;
        }
    }
}